#!/usr/bin/sed -f
#
# utfasc..utf8.sed
#
# Translate escaped unicode text (ASCII \uXXXX) into UTF8.
# This will always be incomplete, but may be good enough for the common cases.
#
# This sed script belongs to utf8..latin1 (but also works on its own).
# (May need to run with LC_CTYPE=C to avoid hangs or errors.)
#
# Copyright (C) by Volker Kuhlmann
#               http://volker.top.geek.nz/contact.html
# All rights reserved.
# Released under the terms of the GNU General Public License (GPL) Version 2.
# See http://www.gnu.org/ for details.
#
# Volker Kuhlmann
#   1.0, 31 May 2013
#   1.1, 08 Jun 2013  Added U+2028, U+2029.
#   1.2, 26 Jun 2013  Added more dashes, apostrophes, quote marks.
#   1.3, 03 Jul 2013  Added (C), (R), (TM).
#   1.3.2, 06 Sep 2013	Added U+2193.
#   1.3.3, 17 Oct 2013  Added U+03C0, U+2248.
#   1.3.4, 06 Dec 2013  Added U+FF08, U+FF09, U+20AC.
#   1.3.5, 12 Dec 2013  Added U+25CF, U+FF1A.
#   1.3.6, 03 Nov 2014  Added U+FF3B, U+FF3D.
#   1.3.7, 22 Jan 2015  Added U+2103, U+2109.
#   1.3.8, 27 Feb 2015  Added U+2605, U+2606.
#   1.3.9, 28 Feb 2015  Added the 10 vowels with macrons.
#   1.3.10, 16 Mar 2015  Added U+201E, U+201F.
#   1.4, 21 Aug 2015  Translate all \uXXXX to UTF-8.
#   1.4.2, 06 Dec 2015
#	Remove old wrong comment.
#	Add comment about sed needing to run with LC_CTYPE=C!!
#   1.4.3, 05 Mar 2016  Remove U+FFFC. Fix comment U+201A. Add comment.
#   1.5, 09 Feb 2018
#	Also translate \uXXXXXXXX to UTF-8.
#	Accept \u and \U for UTF-ASCII.
#   1.6, 25 Mar 2018
#	Only accept [0-9a-fA-F] a hex character - d'oh!
#

#(Note: can't use latin1 characters in here because they are mutually
# exclusive with UTF8 ones in the same file.)

# Line/paragraph separator: see
# http://www.unicode.org/reports/tr13/tr13-9.html

# Note: Find out the UTF8 bytes from the code point with
#   /bin/echo -en '\xff\x08' | recode utf16..utf8 | hex

# sed command			# code point	UTF8	# ASC	UTF8 hex codes	Name
#				#			#(appr)

s/\\u00a9/\xc2\xa9/g		# U+00A9	©	# (C)	c2 a9
s/\\u00ae/\xc2\xae/g		# U+00AE	®	# (R)	c2 ae

s/\\u0100/\xc4\x80/g		# U+0100	Ā	# A-	c4 80
s/\\u0101/\xc4\x81/g		# U+0101	ā	# a-	c4 81
s/\\u0112/\xc4\x92/g		# U+0112	Ē	# E-	c4 92
s/\\u0113/\xc4\x93/g		# U+0113	ē	# e-	c4 93
s/\\u012a/\xc4\xaa/g		# U+012a	Ī	# I-	c4 aa
s/\\u012b/\xc4\xab/g		# U+012b	ī	# i-	c4 ab
s/\\u014c/\xc5\x8c/g		# U+014c	Ō	# O-	c5 8c
s/\\u014d/\xc5\x8d/g		# U+014d	ō	# o-	c5 8d
s/\\u016a/\xc5\xaa/g		# U+016a	Ū	# U-	c5 aa
s/\\u016b/\xc5\xab/g		# U+016b	ū	# u-	c5 ab

s/\\u03c0/\xcf\x80/g		# U+03C0	π	# pi	cf 80

s/\\u2010/\xe2\x80\x90/g	# U+2010	‐	# -	e2 80 90
s/\\u2011/\xe2\x80\x91/g	# U+2011	‑	# -	e2 80 91
s/\\u2012/\xe2\x80\x92/g	# U+2012	‒	# -	e2 80 92
s/\\u2013/\xe2\x80\x93/g	# U+2013	–	# --	e2 80 93
s/\\u2014/\xe2\x80\x94/g	# U+2014	—	# ---	e2 80 94
s/\\u2015/\xe2\x80\x95/g	# U+2015	―	# ---	e2 80 95

s/\\u2018/\xe2\x80\x98/g	# U+2018	‘	# (`)	e2 80 98
s/\\u2019/\xe2\x80\x99/g	# U+2019	’	# (')	e2 80 99
s/\\u201a/\xe2\x80\x9a/g	# U+201a	‚	# (,)	e2 80 9a
s/\\u201b/\xe2\x80\x9b/g	# U+201b	‛	# (')	e2 80 9b
s/\\u201c/\xe2\x80\x9c/g	# U+201c	“	# ``	e2 80 9c
s/\\u201d/\xe2\x80\x9d/g	# U+201d	”	# ''	e2 80 9d
s/\\u201e/\xe2\x80\x9e/g	# U+201e	„	# ,,	e2 80 9e
s/\\u201f/\xe2\x80\x9f/g	# U+201f	‟	# ``	e2 80 9f

s/\\u2022/\xe2\x80\xa2/g	# U+2022	•	# *	e2 80 a2
s/\\u2024/\xe2\x80\xa4/g	# U+2024	․	# .	e2 80 a4
s/\\u2026/\xe2\x80\xa6/g	# U+2026	…	# ...	e2 80 a6

s/\\u2028/\xe2\x80\xa8/g	# U+2028 LS		# NL	e2 80 a8
s/\\u2029/\xe2\x80\xa9/g	# U+2029 PS		# NP	e2 80 a9
s/\\u202f/\xe2\x80\xaf/g	# U+202F NNBSP	 	# SP	e2 80 af

s/\\u20ac/\xe2\x82\xac/g	# U+20AC	€	# EUR	e2 82 ac

s/\\u2103/\xe2\x84\x83/g	# U+2103 degC	℃	# degC	e2 84 83
s/\\u2109/\xe2\x84\x89/g	# U+2109 degF	℉	# degF	e2 84 89

s/\\u2122/\xe2\x84\xa2/g	# U+2122 TM	™	# TM	e2 84 a2

s/\\u2193/\xe2\x86\x93/g	# U+2193	↓	#	e2 86 93

s/\\u2248/\xe2\x89\x88/g	# U+2248	≈	# appr	e2 89 88

s/\\u25cf/\xe2\x97\x8f/g	# U+25CF	●	# *	e2 97 8f	Black circle

s/\\u2605/\xe2\x98\x85/g	# U+2605	★	# star	e2 98 85	Black star
s/\\u2606/\xe2\x98\x85/g	# U+2606	☆	# star	e2 98 86	White star

s/\\uff08/\xef\xbc\x88/g	# U+FF08	（	# (	ef bc 88
s/\\uff09/\xef\xbc\x89/g	# U+FF09	）	# )	ef bc 89
s/\\uff1a/\xef\xbc\x9a/g	# U+FF1A	：	# :	ef bc 9a
s/\\uff3b/\xef\xbc\xbb/g	# U+FF3B	［	# [	ef bc bb
s/\\uff3d/\xef\xbc\xbd/g	# U+FF3D	］	# ]	ef bc bd

# Remove object replacement character.
s/\\ufffc//g			# U+FFFC	￼	# 	ef bf bc


## Translate what is left with a shell script.
# Not fast or efficient, but should work and be fast enough for small files.
# Requires GNU sed!
# Inline shell script, using iconv.

# Fail if not GNU sed 4.0+.
v 4.0

# WARNING:
#    GNU sed version 4.2.1
#    sed (GNU sed) 4.2.2
# These two sed HANG (infinite loop) on the following input:
#   'A \ufb02ex'
#   0000  41 ef bf bd 20 5c 75 66  62 30 32 65 78           A... \uf b02ex
# UNLESS sed is run with LC_CTYPE=C!!

# Translate all \uXXXXXXXX in this line.
:loop8
/\\[uU][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]/ {
    s:':'\\'':g
    s:^\(.*\)\\[uU]\([0-9a-fA-F][0-9a-fA-F]\)\([0-9a-fA-F][0-9a-fA-F]\)\([0-9a-fA-F][0-9a-fA-F]\)\([0-9a-fA-F][0-9a-fA-F]\)\(.*\)$:echo -n '\1'; echo -en '\\x\2\\x\3\\x\4\\x\5' | iconv -f utf32be -t utf8; echo -n '\6':e
    b loop8
}
# Translate all \uXXXX in this line.
:loop4
/\\[uU][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F]/ {
    # Replace all single-quotes with string suitable for echo command.
    s:':'\\'':g
    
    # Replace last occurrence of \uXXXX.
    # (The first .* is greedy and a 'g' flag to s-cmd does nothing.)
    s:^\(.*\)\\[uU]\([0-9a-fA-F][0-9a-fA-F]\)\([0-9a-fA-F][0-9a-fA-F]\)\(.*\)$:echo -n '\1'; echo -en '\\x\2\\x\3' | iconv -f utf16be -t utf8; echo -n '\4':e
    
    # Repeat to replace remaining occurrences of \uXXXX.
    b loop4
}
